static char *fastcopy_id = 
	"@(#)Copyright (C) H.Shirouzu 2004-2007   fastcopy.cpp	Ver1.63";
/* ========================================================================
	Project  Name			: Fast Copy file and directory
	Create					: 2004-09-15(Wed)
	Update					: 2007-07-15(Thu)
	Copyright				: H.Shirouzu
	Reference				: 
	======================================================================== */

#include "fastcopy.h"
#include <stdarg.h>
#include <stddef.h>
#include <process.h>
#include <stdio.h>

void *NTFS_STR_V;		// "NTFS"
void *FMT_RENAME_V;		// ".%03d"
void *FMT_LIST1_V;		//
void *FMT_LIST2_V;		//

#define STRMID_OFFSET offsetof(WIN32_STREAM_ID, cStreamName)
#define MAX_ALTSTREAM 1000

BOOL SetPrivilege(LPTSTR pszPrivilege, BOOL bEnable)
{
    HANDLE           hToken;
    TOKEN_PRIVILEGES tp;

    if (!::OpenProcessToken(::GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
        return FALSE;

    if (!::LookupPrivilegeValue(NULL, pszPrivilege, &tp.Privileges[0].Luid))
        return FALSE;

    tp.PrivilegeCount = 1;

    if (bEnable)
         tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
    else
         tp.Privileges[0].Attributes = 0;

    if (!::AdjustTokenPrivileges(hToken, FALSE, &tp, 0, (PTOKEN_PRIVILEGES)NULL, 0))
         return FALSE;

    if (!::CloseHandle(hToken))
         return FALSE;

    return TRUE;
}

/*=========================================================================
  NX F FastCopy
  T  v F }`Xbhȃt@CRs[ǗNX
     F 
     F 
=========================================================================*/
FastCopy::FastCopy()
{
	hReadThread = hWriteThread = NULL;

	SetPrivilege(SE_BACKUP_NAME, TRUE);
	SetPrivilege(SE_RESTORE_NAME, TRUE);

	::InitializeCriticalSection(&errCs);
	::InitializeCriticalSection(&listCs);

	hRunMutex = ::CreateMutex(NULL, FALSE, FASTCOPY_MUTEX);

	if (IS_WINNT_V) {
		src = new WCHAR [MAX_PATHLEN_V + MAX_PATH];
		dst = new WCHAR [MAX_PATHLEN_V + MAX_PATH];
		confirmDst = new WCHAR [MAX_PATHLEN_V + MAX_PATH];
		NTFS_STR_V		= L"NTFS";
		FMT_INT_STR_V	= L"%d";
		FMT_RENAME_V	= L"%.*s(%d)%s";
		FMT_LIST1_V		= L"%c %s%s%s";
		FMT_LIST2_V		= L"\r\n%c %s%s%s";
	}
	else {
		src = new char [MAX_PATHLEN_V + MAX_PATH_EX];
		dst = new char [MAX_PATHLEN_V + MAX_PATH_EX];
		confirmDst = new char [MAX_PATHLEN_V + MAX_PATH_EX];
		NTFS_STR_V		= "NTFS";
		FMT_INT_STR_V	= "%d";
		FMT_RENAME_V	= "%.*s(%d)%s";
		FMT_LIST1_V		= "%c %s%s%s";
		FMT_LIST2_V		= "\r\n%c %s%s%s";
	}
	maxStatSize = (MAX_PATH * CHAR_LEN_V) * 2 + offsetof(FileStat, cFileName) + 8;
	hashVal.Init(MAX_PATH * CHAR_LEN_V);
	waitTick = 0;
}

FastCopy::~FastCopy()
{
	delete [] confirmDst;
	delete [] dst;
	delete [] src;

	::CloseHandle(hRunMutex);

	::DeleteCriticalSection(&listCs);
	::DeleteCriticalSection(&errCs);
}

BOOL FastCopy::GetRootDir(const void *path, void *root_dir)
{
	if (GetChar(path, 0) == '\\') {	// "\\server\volname\" 4ڂ \ 
		DWORD	ch;
		int		backslash_cnt = 0, offset;

		for (offset=0; (ch = GetChar(path, offset)) != 0 && backslash_cnt < 4; offset++) {
			if (ch == '\\')
				backslash_cnt++;
		}
		memcpy(root_dir, path, offset * CHAR_LEN_V);
		if (backslash_cnt < 4)					// 4 \ Ȃꍇ́A \ t^
			SetChar(root_dir, offset++, '\\');	// i\\server\volume Ȃǁj
		SetChar(root_dir, offset, 0);	// NULL terminate
	}
	else {	// "C:\" 
		memcpy(root_dir, path, 3 * CHAR_LEN_V);
		SetChar(root_dir, 3, 0);	// NULL terminate
	}
	return	TRUE;
}

FastCopy::FsType FastCopy::GetFsType(const void *root_dir)
{
	if (GetDriveTypeV(root_dir) == DRIVE_REMOTE)
		return	FSTYPE_NETWORK;

	DWORD	serial, max_fname, fs_flags;
	WCHAR	vol[MAX_PATH], fs[MAX_PATH];

	if (GetVolumeInformationV(root_dir, vol, MAX_PATH, &serial, &max_fname, &fs_flags, fs, MAX_PATH) == FALSE)
		return	ConfirmErr("GetVolumeInformation", root_dir, FALSE), FSTYPE_NONE;

	return	lstrcmpiV(fs, NTFS_STR_V) == 0 ? FSTYPE_NTFS : FSTYPE_FAT;
}

int FastCopy::GetSectorSize(const void *root_dir)
{
	DWORD	spc, bps, fc, cl;

	if (GetDiskFreeSpaceV(root_dir, &spc, &bps, &fc, &cl) == FALSE) {
//		if (IS_WINNT_V)
//			ConfirmErr("GetDiskFreeSpace", root_dir, FALSE);
		return	OPT_SECTOR_SIZE;
	}
	return	bps;
}

BOOL FastCopy::IsSameDrive(const void *root1, const void *root2)
{
	if (GetChar(root1, 1) == ':' && GetChar(root2, 1) == ':')
		return	driveMng.IsSameDrive(GetChar(root1, 0), GetChar(root2, 0));

	if (GetChar(root1, 1) != GetChar(root2, 1) || GetDriveTypeV(root1) != GetDriveTypeV(root2))
		return	FALSE;

	return	lstrcmpiV(root1, root2) == 0 ? TRUE : FALSE;
}

int FastCopy::MakeUnlimitedPath(WCHAR *buf)
{
	int		prefix_len;
	WCHAR	*prefix;
	BOOL	isUNC = (*buf == '\\') ? TRUE : FALSE;

	prefix		= isUNC ? PATH_UNC_PREFIX : PATH_LOCAL_PREFIX;
	prefix_len	= isUNC ? PATH_UNC_PREFIX_LEN : PATH_LOCAL_PREFIX_LEN;

	// (isUNC ? 1 : 0) ... PATH_UNC_PREFIX ̏ꍇA\\server -> \\?\UNC\server 
	//  ɂ邽߁A\\server ̓ \ ׂB
	memmove(buf + prefix_len - (isUNC ? 1 : 0), buf, (strlenV(buf) + 1) * CHAR_LEN_V);
	memcpy(buf, prefix, prefix_len * CHAR_LEN_V);
	return	prefix_len;
}

BOOL FastCopy::InitDstPath(void)
{
	DWORD	attr;
	WCHAR	wbuf[MAX_PATH_EX];
	void	*buf = wbuf, *fname = NULL;
	const void	*org_path = dstArray.Path(0), *dst_root;

	// dst ̊mF/H
	if (GetChar(org_path, 1) == ':' && GetChar(org_path, 2) != '\\')
		return	ConfirmErr(GetLoadStr(IDS_BACKSLASHERR), org_path, FALSE), FALSE;

	if (GetFullPathNameV(org_path, MAX_PATHLEN_V, dst, &fname) == 0)
		return	ConfirmErr("GetFullPathName", org_path, FALSE), FALSE;

	GetRootDir(dst, buf);
	dstArray.RegisterPath(buf);
	dst_root = dstArray.Path(dstArray.Num() -1);

	attr = GetFileAttributesV(dst);

	if ((attr = GetFileAttributesV(dst)) == 0xffffffff) {
		if (info.flags & LISTING_ONLY)
			info.overWrite = BY_ALWAYS;	// dest ݂Ȃ߁A̕KvȂ
		else {
			CreateDirectoryV(dst, NULL);
			if ((attr = GetFileAttributesV(dst)) == 0xffffffff)
				return	ConfirmErr("CreateDirectory", dst, FALSE), FALSE;
		}
	}
	if ((attr & FILE_ATTRIBUTE_DIRECTORY) == 0)
		return	ConfirmErr("Not a directory", dst, FALSE), FALSE;

	strcpyV(buf, dst);
	MakePathV(dst, buf, EMPTY_STR_V);
	// srĉRs[邩idst \ Ă or wj
	isExtendDir = strcmpV(buf, dst) == 0 || srcArray.Num() > 1 ? TRUE : FALSE;
	dstPrefixLen = IS_WINNT_V ? MakeUnlimitedPath((WCHAR *)dst) : 0;
	dstBaseLen = strlenV(dst);

	// dst t@CVXe擾
	dstSectorSize = GetSectorSize(dst_root);
	dstFsType = GetFsType(dst_root);
	nbMinSize = dstFsType == FSTYPE_NTFS ? info.nbMinSizeNtfs : info.nbMinSizeFat;

	// ő]TCY
	maxTransSize = isSameDrv ? info.bufSize : info.bufSize / 2;
	maxTransSize = min(info.maxTransSize, maxTransSize);

	// Rs[pdstt@CmF
	strcpyV(confirmDst, dst);

	return	TRUE;
}

BOOL FastCopy::InitSrcPath(int idx)
{
	DWORD		attr;
	BYTE		src_root_cur[MAX_PATH];
	WCHAR		wbuf[MAX_PATH_EX];
	void		*buf = wbuf, *fname = NULL;
	const void	*dst_root = dstArray.Path(dstArray.Num() -1);
	const void	*org_path = srcArray.Path(idx);

	// src ̊mF/H
	if (GetChar(org_path, 1) == ':' && GetChar(org_path, 2) != '\\')
		return	ConfirmErr(GetLoadStr(IDS_BACKSLASHERR), org_path), FALSE;

	if (GetFullPathNameV(org_path, MAX_PATHLEN_V, src, &fname) == 0)
		return	ConfirmErr("GetFullPathName", org_path), FALSE;
	GetRootDir(src, src_root_cur);

	if ((attr = GetFileAttributesV(src)) == 0xffffffff) {
		isMetaSrc = TRUE;
	}
	else if (attr & FILE_ATTRIBUTE_DIRECTORY) {
		isMetaSrc = FALSE;
		// efBNĝRs[ȂꍇA\* t^
		strcpyV(buf, src);
		MakePathV(src, buf, ASTERISK_V);
		if (strcmpV(buf, src_root_cur) == 0 || !isExtendDir)
			isMetaSrc = TRUE;
		else
			SetChar(src, strlenV(src) - 2, 0);	//  \* tȂ
	}
	srcPrefixLen = IS_WINNT_V ? MakeUnlimitedPath((WCHAR *)src) : 0;

	if (GetFullPathNameV(src, MAX_PATH_EX, buf, &fname) == 0 || fname == NULL)
		return	ConfirmErr("GetFullPathName2", MakeAddr(src, srcPrefixLen)), FALSE;

	// mFpdst
	strcpyV(MakeAddr(confirmDst, dstBaseLen), fname);

	// pXłȂƂ̊mF
	if (lstrcmpiV(buf, confirmDst) == 0) {
		if (info.mode != DIFFCP_MODE || (info.flags & SAMEDIR_RENAME) == 0)
			return	ConfirmErr(GetLoadStr(IDS_SAMEPATHERR), MakeAddr(confirmDst, dstBaseLen), FALSE), FALSE;
		strcpyV(MakeAddr(confirmDst, dstBaseLen), ASTERISK_V);
		isRename = TRUE;
	}
	else
		isRename = FALSE;

	if (info.mode == MOVE_MODE && (attr & FILE_ATTRIBUTE_DIRECTORY)) {	// eqւ̈ړ͔F߂Ȃ
		int	end_offset = 0;
		if (GetChar(fname, 0) == '*' || attr == 0xffffffff) {
			SetChar(fname, 0, 0);
			end_offset = 1;
		}
		int		len = IS_WINNT_V ? strlenV(buf) : ::lstrlenA((char *)buf);
		if (srcBaseLen <= dstBaseLen && strnicmpV(buf, confirmDst, len) == 0) {
			DWORD ch = GetChar(confirmDst, len - end_offset);
			if (ch == 0 || ch == '\\') {
				return	ConfirmErr(GetLoadStr(IDS_PARENTPATHERR), MakeAddr(buf, srcPrefixLen), FALSE), FALSE;
			}
		}
	}

	SetChar(fname, 0, 0);
	srcBaseLen = strlenV(buf);

	// src t@CVXe擾
	if (lstrcmpiV(src_root_cur, src_root)) {
		srcSectorSize = GetSectorSize(src_root_cur);
		srcFsType = GetFsType(src_root_cur);

		sectorSize = max(srcSectorSize, dstSectorSize);		// 傫قɍ킹
		sectorSize = max(sectorSize, MIN_SECTOR_SIZE);

		// ꕨhCuǂ̒
		if (info.flags & FIX_SAMEDISK)
			isSameDrv = TRUE;
		else if (info.flags & FIX_DIFFDISK)
			isSameDrv = FALSE;
		else
			isSameDrv = IsSameDrive(src_root_cur, dst_root);
		if (info.mode == MOVE_MODE)
			isSameVol = strcmpV(src_root_cur, dst_root);
	}

	enableAcl = (info.flags & WITH_ACL) && srcFsType != FSTYPE_FAT && dstFsType != FSTYPE_FAT;
	enableStream = (info.flags & WITH_ALTSTREAM) && srcFsType != FSTYPE_FAT && dstFsType != FSTYPE_FAT;

	strcpyV(src_root, src_root_cur);

	return	TRUE;
}

BOOL FastCopy::InitDeletePath(int idx)
{
	DWORD		attr;
	WCHAR		wbuf[MAX_PATH_EX];
	void		*buf = wbuf, *fname = NULL;
	const void	*org_path = srcArray.Path(idx);
	BYTE		dst_root[MAX_PATH];

	// delete p path ̊mF/H
	if (GetChar(org_path, 1) == ':' && GetChar(org_path, 2) != '\\')
		return	ConfirmErr(GetLoadStr(IDS_BACKSLASHERR), org_path), FALSE;

	if (GetFullPathNameV(org_path, MAX_PATHLEN_V, dst, &fname) == 0)
		return	ConfirmErr("GetFullPathName", org_path), FALSE;

	attr = GetFileAttributesV(dst);

	if (attr != 0xffffffff && (attr & FILE_ATTRIBUTE_DIRECTORY) || (info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA))) {
		GetRootDir(dst, dst_root);
		if (info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA)) {
			dstSectorSize = GetSectorSize(dst_root);
			dstFsType = GetFsType(dst_root);
			nbMinSize = dstFsType == FSTYPE_NTFS ? info.nbMinSizeNtfs : info.nbMinSizeFat;
		}
	}

	if (attr == 0xffffffff) {
		isMetaSrc = TRUE;
	}
	else if (attr & FILE_ATTRIBUTE_DIRECTORY) {
		isMetaSrc = FALSE;
		strcpyV(buf, dst);
		// root_dir ͖ "\*" t^AȊO͖ "\"폜
		MakePathV(dst, buf, ASTERISK_V);
		if (lstrcmpiV(buf, dst_root) == 0)
			isMetaSrc = TRUE;
		else
			SetChar(dst, strlenV(dst) - 2, 0);
	}
	dstPrefixLen = IS_WINNT_V ? MakeUnlimitedPath((WCHAR *)dst) : 0;

	if (GetFullPathNameV(dst, MAX_PATH_EX, buf, &fname) == 0 || fname == NULL)
		return	ConfirmErr("GetFullPathName2", MakeAddr(dst, dstPrefixLen)), FALSE;
	SetChar(fname, 0, 0);
	dstBaseLen = strlenV(buf);

	if (info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA)) {
		strcpyV(confirmDst, dst);	// for renaming before deleting
	}

	return	TRUE;
}

BOOL FastCopy::RegisterInfo(const PathArray *_srcArray, const PathArray *_dstArray, Info *_info, const PathArray *_includeArray, const PathArray *_excludeArray)
{
	isAbort = FALSE;
	isRename = FALSE;
	isFilter = FALSE;
	endTick = 0;

	info = *_info;
	SetChar(src_root, 0, 0);

	if (info.flags & LISTING_ONLY)
		info.flags &= ~PRE_SEARCH;

	// filter
	PathArray	pathArray[] = { *_includeArray, *_excludeArray };
	void		*path = NULL;

	for (int kind=0; kind < MAX_KIND_EXP; kind++) {
		for (int ftype=0; ftype < MAX_FTYPE_EXP; ftype++)
			regExp[kind][ftype].Init();

		for (int idx=0; idx < pathArray[kind].Num(); idx++) {
			int		len = lstrlenV(path = pathArray[kind].Path(idx)), targ = FILE_EXP;

			if (lGetCharV(path, len -1) == '\\') {
				lSetCharV(path, len -1, 0);
				targ = DIR_EXP;
			}
			if (!regExp[kind][targ].RegisterWildCard(path, RegExp::CASE_INSENSE))
				return	ConfirmErr("Bad or Too long windcard string", path, FALSE), FALSE;
		}
	}
	if (path)
		isFilter = TRUE;

	if ((info.flags & LISTING_ONLY) == 0 &&
		(info.mode != DELETE_MODE || (info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA))) &&
		(info.bufSize > MAX_BUF || info.bufSize < MIN_BUF * 2))
		return	ConfirmErr("Too large or small Main Buffer.", NULL, FALSE), FALSE;

	// command
	if (info.mode == DELETE_MODE) {
		srcArray = *_srcArray;
		if (InitDeletePath(0) == FALSE)
			return	FALSE;
	}
	else {
		srcArray = *_srcArray;
		dstArray = *_dstArray;

		if (InitDstPath() == FALSE)
			return	FALSE;
		if (InitSrcPath(0) == FALSE)
			return	FALSE;
	}
	_info->isRenameMode = isRename;

	return	!isAbort;
}

BOOL FastCopy::TakeExclusivePriv(void)
{
	return	::WaitForSingleObject(hRunMutex, 0) == WAIT_OBJECT_0 ? TRUE : FALSE;
}


BOOL FastCopy::Start(void)
{
	u_int	id;
	int		allocSize = (info.flags & LISTING_ONLY) ? MAX_LIST_BUF : info.bufSize+PAGE_SIZE;
	memset(&total, 0, sizeof(total));
	if (info.flags & PRE_SEARCH)
		total.isPreSearch = TRUE;

	isAbort = FALSE;
	writeReq = NULL;
	isSuspend = FALSE;
	readDestStatQueue = FALSE;
	nextFileID = 1;
	errFileID = 0;
	openFiles = NULL;

	cv.Initialize(2);
	readReqList.Init();
	writeReqList.Init();

	errBuf.AllocBuf(MIN_ERR_BUF, MAX_ERR_BUF);

	if (info.flags & LISTING_ONLY)
		listBuf.AllocBuf(MIN_PUTLIST_BUF, MAX_PUTLIST_BUF);

	if (info.mode == DELETE_MODE) {
		if ((info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA)) && (info.flags & LISTING_ONLY) == 0) {
			if (mainBuf.AllocBuf(allocSize) == FALSE) {
				ConfirmErr("Can't alloc memory(delBuf)", NULL, FALSE);
				goto ERR;
			}
			SetupRandomDataBuf();
		}
		startTick = ::GetTickCount();
		if (!(hReadThread = (HANDLE)_beginthreadex(0, 0, FastCopy::DeleteThread, this, 0, &id)))
			goto ERR;
		return	TRUE;
	}

	openFiles = new FileStat *[info.maxOpenFiles + MAX_ALTSTREAM]; /* for Alternate Stream */
	openFilesCnt = 0;

	// src/dst dir-entry/attr pobt@m
	fileStatBuf.AllocBuf(MIN_ATTR_BUF, info.maxAttrSize);
	dirStatBuf.AllocBuf(MIN_ATTR_BUF, info.maxDirSize);
	dstStatBuf.AllocBuf(MIN_ATTR_BUF, info.maxAttrSize);
	dstStatIdxBuf.AllocBuf(MIN_ATTRIDX_BUF, MAX_ATTRIDX_BUF(info.maxAttrSize));
	if (info.flags & SKIP_EMPTYDIR)
		mkdirQueueBuf.AllocBuf(MIN_MKDIRQUEUE_BUF, MAX_MKDIRQUEUE_BUF);
	mkdirAclBuf.AllocBuf(MIN_MKDIRACL_BUF, MAX_MKDIRACL_BUF);

	// COobt@m
	if (mainBuf.AllocBuf(allocSize) == FALSE && (info.flags & LISTING_ONLY) == 0) {
		ConfirmErr("Can't alloc memory(mainBuf)", NULL, FALSE);
		goto ERR;
	}

	// Oobt@pItZbg
	usedOffset = freeOffset = mainBuf.Buf();

	if (IS_WINNT_V && info.isPhysLock) {
		::SetProcessWorkingSetSize(::GetCurrentProcess(), mainBuf.Size() + APP_MEMSIZE, mainBuf.Size() + APP_MEMSIZE);
		mainBuf.LockBuf();
		fileStatBuf.LockBuf();
		dirStatBuf.LockBuf();
		dstStatBuf.LockBuf();
		dstStatIdxBuf.LockBuf();
	}

	startTick = ::GetTickCount();

	if (!(hReadThread = (HANDLE)_beginthreadex(0, 0, FastCopy::ReadThread, this, 0, &id))
	|| !(hWriteThread = (HANDLE)_beginthreadex(0, 0, FastCopy::WriteThread, this, 0, &id)))
		goto ERR;
	return	TRUE;

ERR:
	End();
	return	FALSE;
}

/*=========================================================================
     F ReadThread
  T  v F Read 
     F 
     F 
=========================================================================*/
unsigned WINAPI FastCopy::ReadThread(void *fastCopyObj)
{
	return	((FastCopy *)fastCopyObj)->ReadThreadCore();
}

BOOL FastCopy::ReadThreadCore(void)
{
	BOOL	isSameDrvOld;
	int		done_cnt = 0;

	if (info.flags & PRE_SEARCH)
		PreSearch();

	for (int i=0; i < srcArray.Num(); i++) {
		if (InitSrcPath(i)) {
			if (done_cnt >= 1 && isSameDrvOld != isSameDrv)
				ChangeToWriteMode();
			ReadProc(srcBaseLen, info.overWrite == BY_ALWAYS ? FALSE : TRUE);
			isSameDrvOld = isSameDrv;
			done_cnt++;
		}
		if (isAbort)
			break;
	}
	SendRequest(REQ_EOF);
	ChangeToWriteMode();
	return	TRUE;
}

BOOL FastCopy::PreSearch(void)
{
	BOOL	is_delete = info.mode == DELETE_MODE;
	BOOL	(FastCopy::*InitPath)(int) = is_delete ? InitDeletePath : InitSrcPath;
	void	*&path = is_delete ? dst : src;
	int		&prefix_len = is_delete ? dstPrefixLen : srcPrefixLen;
	int		&base_len = is_delete ? dstBaseLen : srcBaseLen;
	BOOL	ret = TRUE;

	for (int i=0; i < srcArray.Num(); i++) {
		if ((this->*InitPath)(i)) {
			if (!PreSearchProc(path, prefix_len, base_len))
				ret = FALSE;
		}
		if (isAbort)
			break;
	}
	total.isPreSearch = FALSE;
	startTick = ::GetTickCount();

	return	ret && !isAbort;
}

BOOL FastCopy::FilterCheck(const void *path, const void *fname, DWORD attr)
{
	int	targ = (attr & FILE_ATTRIBUTE_DIRECTORY) ? DIR_EXP : FILE_EXP;

	if (regExp[EXC_EXP][targ].IsMatch(fname))
		return	FALSE;

	if (!regExp[INC_EXP][targ].IsRegistered())
		return	TRUE;

	if (regExp[INC_EXP][targ].IsMatch(fname))
		return	TRUE;

	return	FALSE;
}

BOOL FastCopy::PreSearchProc(void *path, int prefix_len, int dir_len)
{
	HANDLE		hDir;
	BOOL		ret = TRUE;
	WIN32_FIND_DATAW fdat;

	if (waitTick) Wait(1);

	if ((hDir = FindFirstFileV(path, &fdat)) == INVALID_HANDLE_VALUE) {
		return	ConfirmErr("FindFirstFile(pre)", MakeAddr(path, prefix_len)), FALSE;
	}

	do {
		if (IsParentOrSelfDirs(fdat.cFileName))
			continue;

		// src fBNĝɑ΂ẮAtB^ΏۂɂȂ
		if (isFilter && (dir_len != srcBaseLen || isMetaSrc) && !FilterCheck(path, fdat.cFileName, fdat.dwFileAttributes))
			continue;

		if (fdat.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			total.preDirs++;
			int len = sprintfV(MakeAddr(path, dir_len), FMT_CAT_ASTER_V, fdat.cFileName) -1;
			ret = PreSearchProc(path, prefix_len, dir_len + len);
		}
		else {
			total.preFiles++;
			total.preTrans += (((_int64)fdat.nFileSizeHigh << 32) + fdat.nFileSizeLow);
		}
	}
	while (!isAbort && FindNextFileV(hDir, &fdat));

	if (!isAbort && ret && ::GetLastError() != ERROR_NO_MORE_FILES) {
		ret = FALSE;
		ConfirmErr("FindNextFile(pre)", MakeAddr(path, prefix_len));
	}

	::FindClose(hDir);

	return	ret && !isAbort;
}

BOOL FastCopy::PutList(void *path, DWORD opt)
{
	BOOL	add_backslash = GetChar(path, 0) == '\\' && GetChar(path, 1) != '\\';

	::EnterCriticalSection(&listCs);

	if (listBuf.RemainSize() >= MAX_PATHLEN_V || listBuf.Grow(MIN_PUTLIST_BUF)) {
		int len = sprintfV(listBuf.Buf() + listBuf.UsedSize(), listBuf.UsedSize() ? FMT_LIST2_V : FMT_LIST1_V, (opt & PL_DELETE) ? '-' : '+', add_backslash ? BACK_SLASH_V : EMPTY_STR_V, path, (opt & PL_DIRECTORY) ? L"\\" : L"");

		listBuf.AddUsedSize(len * CHAR_LEN_V);
	}

	::LeaveCriticalSection(&listCs);
	return	TRUE;
}

BOOL FastCopy::ReadProc(int dir_len, BOOL confirm_dir)
{
	BOOL		ret = TRUE;
	FileStat	*srcStat, *statEnd;
	FileStat	*dstStat = NULL;
	int			new_dir_len, curDirStatSize, confirm_len;
	BOOL		is_rename_local = isRename;
	BOOL		confirm_dir_local = confirm_dir || is_rename_local;

	isRename = FALSE;	// top level ̂݌ʂo

	if (waitTick) Wait(1);

	// Jg̃TCYۑ
	curDirStatSize = dirStatBuf.UsedSize();

	if (confirm_dir_local && !isSameDrv)
		ReadDestStatRequest();

	// fBNgGgɂׂēǂݎ
	ret = ReadDirEntry(dir_len, confirm_dir_local);

	if (confirm_dir_local && (isSameDrv ? ReadDestStat() : WaitReadDestStat()) == FALSE || isAbort || !ret)
		return	FALSE;

	// t@Cɏ
	statEnd = (FileStat *)(fileStatBuf.Buf() + fileStatBuf.UsedSize());
	for (srcStat = (FileStat *)fileStatBuf.Buf(); srcStat < statEnd; srcStat = (FileStat *)((BYTE *)srcStat + srcStat->size)) {
		if (confirm_dir_local)
			dstStat = hash.Search(srcStat->upperName, srcStat->hashVal);

		if (dstStat) {
			if (is_rename_local) {
				SetRenameCount(srcStat);
			}
			else {
				dstStat->isExists = TRUE;
				if (IsOverWriteFile(srcStat, dstStat) == FALSE) {
					total.skipFiles++;
					total.skipTrans += dstStat->FileSize();
					continue;
				}
			}
		}
		total.readFiles++;

		if (OpenFileProc(srcStat, dir_len) == FALSE || srcFsType == FSTYPE_NETWORK || waitTick && openFilesCnt >= 10 || openFilesCnt >= info.maxOpenFiles) {
			ReadMultiFilesProc();
			CloseMultiFilesProc();
		}
		if (isAbort)
			goto END;
	}

	ReadMultiFilesProc();
	CloseMultiFilesProc();
	if (isAbort)
		goto END;

	statEnd = (FileStat *)(dirStatBuf.Buf() + dirStatBuf.UsedSize());

	// fBNg̑݊mF
	if (confirm_dir_local) {
		confirm_len = dir_len + (dstBaseLen - srcBaseLen);

		for (srcStat = (FileStat *)(dirStatBuf.Buf() + curDirStatSize); srcStat < statEnd; srcStat = (FileStat *)((BYTE *)srcStat + srcStat->size)) {
			if ((dstStat = hash.Search(srcStat->upperName, srcStat->hashVal)) != NULL) {
				if (is_rename_local)
					SetRenameCount(srcStat);
				else
					srcStat->isExists = dstStat->isExists = TRUE;
			}
			else
				srcStat->isExists = FALSE;
		}
	}

	// SYNC[h̏ꍇARs[ɖt@C폜
	if (confirm_dir_local && info.mode == SYNCCP_MODE) {
		int		max = dstStatIdxBuf.UsedSize() / sizeof(FileStat *);
		for (int i=0; i < max; i++) {
			if ((dstStat = ((FileStat **)dstStatIdxBuf.Buf())[i])->isExists)
				continue;

			if (isFilter && !FilterCheck(confirmDst, dstStat->cFileName, dstStat->dwFileAttributes))
				continue;

			if (isSameDrv) {
				if (dstStat->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
					ret = DeleteDirProc(confirmDst, confirm_len, dstStat->cFileName, dstStat);
				else
					ret = DeleteFileProc(confirmDst, confirm_len, dstStat->cFileName, dstStat);
			}
			else {
				SendRequest(DELETE_FILES, 0, dstStat);
			}
			if (isAbort)
				goto END;
		}
	}

	// fBNg
	for (srcStat = (FileStat *)(dirStatBuf.Buf() + curDirStatSize); srcStat < statEnd; srcStat = (FileStat *)((BYTE *)srcStat + srcStat->size)) {
		total.readDirs++;
		ReqBuf	req_buf = {0};
		new_dir_len = dir_len + sprintfV(MakeAddr(src, dir_len), FMT_STR_V, srcStat->cFileName);
		if (confirm_dir && srcStat->isExists)
			sprintfV(MakeAddr(confirmDst, confirm_len), FMT_CAT_ASTER_V, srcStat->cFileName);

		if (enableAcl && (info.flags & LISTING_ONLY) == 0) {
			GetDirAcl(&req_buf, srcStat);
		}
		new_dir_len += sprintfV(MakeAddr(src, new_dir_len), FMT_CAT_ASTER_V, L"") - 1;
		if (SendRequest(confirm_dir && srcStat->isExists ? INTODIR : MKDIR, req_buf.buf ? &req_buf : 0, srcStat), isAbort)
			goto END;

		int	curFiles = total.skipFiles + total.skipDirs + total.readFiles + total.readDirs;
		if (ReadProc(new_dir_len, confirm_dir && srcStat->isExists), isAbort)
			goto END;
		if (total.skipFiles + total.skipDirs + total.readFiles + total.readDirs > curFiles) {
		}
		if (SendRequest(RETURN_PARENT), isAbort)
			goto END;
	}

END:
	// Jg dirpBuf TCY𕜌
	dirStatBuf.SetUsedSize(curDirStatSize);
	return	ret && !isAbort;
}

BOOL FastCopy::GetDirAcl(ReqBuf *req_buf, FileStat *stat)
{
	HANDLE	fh;
	WIN32_STREAM_ID	sid;
	DWORD	size = 0, lowSeek, highSeek;
	void	*context = NULL;
	int		altdata_cnt = 0;
	BYTE	streamName[MAX_PATH * sizeof(WCHAR)];

	if ((fh = CreateFileV(src, GENERIC_READ|READ_CONTROL, 0, 0, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0)) == INVALID_HANDLE_VALUE)
		return	FALSE;

	do {
		if (!::BackupRead(fh, (LPBYTE)&sid, STRMID_OFFSET, &size, FALSE, TRUE, &context) || size != STRMID_OFFSET) {
			DWORD	errCode = ::GetLastError();
			if (errCode != 0 && errCode != ERROR_INVALID_HANDLE && (info.flags & REPORT_ACL_ERROR))
				ConfirmErr("BackupRead(DIR)", src);
			break;
		}
		if (sid.dwStreamNameSize && !::BackupRead(fh, streamName, sid.dwStreamNameSize, &size, FALSE, TRUE, &context)) {
			break;
		}
		if (sid.dwStreamId == BACKUP_SECURITY_DATA) {
			stat->aclSize = sid.Size.LowPart + STRMID_OFFSET;
			if (PrepareReqBuf(offsetof(ReqHeader, stat) + stat->minSize, stat->aclSize, stat->fileID, req_buf)) {
				memcpy(req_buf->buf, &sid, STRMID_OFFSET);
				if (!::BackupRead(fh, req_buf->buf + STRMID_OFFSET, sid.Size.LowPart, &size, FALSE, TRUE, &context) || size <= 0) {
					if (info.flags & REPORT_ACL_ERROR)
						ConfirmErr("BackupRead(DIRACL)", src);
				}
				else stat->acl = req_buf->buf;
			}
			break;
		}
		if ((sid.Size.LowPart || sid.Size.HighPart) && !::BackupSeek(fh, sid.Size.LowPart, sid.Size.HighPart, &lowSeek, &highSeek, &context)) {
			if (info.flags & REPORT_ACL_ERROR)
				ConfirmErr("BackupSeek(DIR)", dst);
			break;
		}
	}
	while (sid.dwStreamId != BACKUP_SECURITY_DATA);

	::BackupWrite(fh, NULL, NULL, NULL, TRUE, TRUE, &context);
	::CloseHandle(fh);
	return	stat->acl ? TRUE : FALSE;
}

int FastCopy::MakeRenameName(void *buf, int count, void *fname)
{
	void	*ext = strrchrV(fname, '.');
	int		body_limit = ext ? DiffLen(ext, fname) : MAX_PATH;

	return	sprintfV(buf, FMT_RENAME_V, body_limit, fname, count, ext ? ext : EMPTY_STR_V);
}

BOOL FastCopy::SetRenameCount(FileStat *stat)
{
	while (1) {
		WCHAR	new_name[MAX_PATH];
		int		len = MakeRenameName(new_name, ++stat->renameCount, stat->upperName) + 1;
		DWORD	hash_val = hashVal.MakeHash(new_name, len * CHAR_LEN_V);
		if (hash.Search(new_name, hash_val) == NULL)
			break;
	}
	return	TRUE;
}

/*
	㏑
*/
BOOL FastCopy::IsOverWriteFile(FileStat *srcStat, FileStat *dstStat)
{
	if (info.overWrite == BY_NAME)
		return	FALSE;

// time stamp monitor
#if 0
	_int64	st = *(_int64 *)&srcStat->ftLastWriteTime;
	_int64	dt = *(_int64 *)&dstStat->ftLastWriteTime;
	if (st != dt && (info.flags & LISTING_ONLY)) {
		::EnterCriticalSection(&listCs);

		int len = sprintfV(listBuf.Buf() + listBuf.UsedSize(), L" src/dst: %I64d / %I64d", st, dt);
		listBuf.AddUsedSize(len * CHAR_LEN_V);

		::LeaveCriticalSection(&listCs);
	}
#endif

	if (info.overWrite == BY_ATTR) {
		// TCYA...
		if (dstStat->FileSize() == srcStat->FileSize()) {
			if (*(_int64 *)&dstStat->ftLastWriteTime == *(_int64 *)&srcStat->ftLastWriteTime) {	// XVtSɓ
				return	FALSE;
			}
			if (srcFsType != FSTYPE_NTFS || dstFsType != FSTYPE_NTFS) {	// ǂ炩 NTFS łȂꍇilbg[NhCu܂ށj
				// Е NTFS łȂꍇA1msec ̌덷͋eŁAriUDF ΍j
				if (*(_int64 *)&dstStat->ftLastWriteTime + 10000 >= *(_int64 *)&srcStat->ftLastWriteTime &&
					*(_int64 *)&dstStat->ftLastWriteTime - 10000 <= *(_int64 *)&srcStat->ftLastWriteTime) {
					return	FALSE;
				}
				// src  dst ̃^CX^v̍ŏPʂ 1 bȏiFAT or SAMBA ̉\jł...
				if ((*(_int64 *)&srcStat->ftLastWriteTime % 10000000) == 0 || (*(_int64 *)&dstStat->ftLastWriteTime % 10000000) == 0) {
					// ^CX^v̍ 2 bȓȂA^CX^vƂ݂ȂāA㏑Ȃ
					if (*(_int64 *)&dstStat->ftLastWriteTime + 20000000 >= *(_int64 *)&srcStat->ftLastWriteTime &&
						*(_int64 *)&dstStat->ftLastWriteTime - 20000000 <= *(_int64 *)&srcStat->ftLastWriteTime)
						return	FALSE;
				}
			}
		}
		return	TRUE;
	}

	if (info.overWrite == BY_LASTEST) {
		// XVt dst ƓÂꍇ͍XVȂ
		if (*(_int64 *)&dstStat->ftLastWriteTime >= *(_int64 *)&srcStat->ftLastWriteTime) {
			return	FALSE;
		}
		if (srcFsType != FSTYPE_NTFS || dstFsType != FSTYPE_NTFS) {	// ǂ炩 NTFS łȂꍇilbg[NhCu܂ށj
			// Е NTFS łȂꍇA1msec ̌덷͋eŁAriUDF ΍j
			if (*(_int64 *)&dstStat->ftLastWriteTime + 10000 >= *(_int64 *)&srcStat->ftLastWriteTime) {
				return	FALSE;
			}
			// src  dst ̃^CX^v̍ŏPʂ 1 bȏiFAT or SAMBA ̉\jł...
			if ((*(_int64 *)&srcStat->ftLastWriteTime % 10000000) == 0 || (*(_int64 *)&dstStat->ftLastWriteTime % 10000000) == 0) {
				// ^CX^v̍ 2 b̃}[WtŁA XVt dst ƓÂꍇ́A㏑Ȃ
				if (*(_int64 *)&dstStat->ftLastWriteTime + 20000000 >= *(_int64 *)&srcStat->ftLastWriteTime)
					return	FALSE;
			}
		}
		return	TRUE;
	}

	if (info.overWrite == BY_ALWAYS)
		return	TRUE;

	return	ConfirmErr("Illegal overwrite mode", 0, FALSE), FALSE;
}

BOOL FastCopy::ReadDirEntry(int dir_len, BOOL confirm_dir)
{
	HANDLE	fh;
	BOOL	ret = TRUE;
	int		len;
	WIN32_FIND_DATAW fdat;

	fileStatBuf.SetUsedSize(0);

	if ((fh = FindFirstFileV(src, &fdat)) == INVALID_HANDLE_VALUE) {
		total.errDirs++;
		return	ConfirmErr("FindFirstFile", MakeAddr(src, srcPrefixLen)), FALSE;
	}
	do {
		if (IsParentOrSelfDirs(fdat.cFileName))
			continue;

		// src fBNĝɑ΂ẮAtB^ΏۂɂȂ
		if (isFilter && (dir_len != srcBaseLen || isMetaSrc) && !FilterCheck(src, fdat.cFileName, fdat.dwFileAttributes))
			continue;

		// fBNgt@C̒~
		if (fdat.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			len = FdatToFileStat(&fdat, (FileStat *)(dirStatBuf.Buf() + dirStatBuf.UsedSize()), confirm_dir);
			dirStatBuf.AddUsedSize(len);
			if (dirStatBuf.RemainSize() <= maxStatSize && dirStatBuf.Grow(MIN_ATTR_BUF) == FALSE) {
				ConfirmErr("Can't alloc memory(dirStatBuf)", NULL, FALSE);
				break;
			}
		}
		else {
			len = FdatToFileStat(&fdat, (FileStat *)(fileStatBuf.Buf() + fileStatBuf.UsedSize()), confirm_dir);
			fileStatBuf.AddUsedSize(len);
			if (fileStatBuf.RemainSize() <= maxStatSize && fileStatBuf.Grow(MIN_ATTR_BUF) == FALSE) {
				ConfirmErr("Can't alloc memory(fileStatBuf)", NULL, FALSE);
				break;
			}
		}
	}
	while (!isAbort && FindNextFileV(fh, &fdat));

	if (!isAbort && ::GetLastError() != ERROR_NO_MORE_FILES) {
		total.errFiles++;
		ret = FALSE;
		ConfirmErr("FindNextFile", MakeAddr(src, srcPrefixLen));
	}
	::FindClose(fh);

	return	ret && !isAbort;
}

BOOL FastCopy::OpenFileProc(FileStat *stat, int dir_len)
{
	DWORD	name_len = strlenV(stat->cFileName);
	memcpy(MakeAddr(src, dir_len), stat->cFileName, ((name_len + 1) * CHAR_LEN_V));

	stat->fileID = nextFileID++;
	openFiles[openFilesCnt++] = stat;

	if (waitTick) Wait((waitTick + 9) / 10);

	BOOL	isBackup = enableAcl || enableStream;

	if (!isBackup && stat->FileSize() == 0 || (info.flags & LISTING_ONLY))
		return	TRUE;

	DWORD	mode = GENERIC_READ;
	DWORD	flg = (info.flags & USE_OSCACHE_READ) ? FILE_FLAG_SEQUENTIAL_SCAN : (FILE_FLAG_NO_BUFFERING|FILE_FLAG_SEQUENTIAL_SCAN);

	if (isBackup) {
		mode |= READ_CONTROL;
		flg  |= FILE_FLAG_BACKUP_SEMANTICS;
	}

	if ((stat->hFile = CreateFileV(src, mode, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, flg, 0)) == INVALID_HANDLE_VALUE) {
		stat->lastError = ::GetLastError();
		return	FALSE;
	}

	return	isBackup ? OpenFileBackupProc(stat, dir_len + name_len) : TRUE;
}

BOOL FastCopy::OpenFileBackupProc(FileStat *stat, int src_len)
{
	WIN32_STREAM_ID	sid;
	BOOL	ret = FALSE;
	DWORD	size = 0;
	DWORD	lowSeek, highSeek;
	void	*context = NULL;
	int		altdata_cnt = 0;
	BYTE	streamName[MAX_PATH * sizeof(WCHAR)];
	DWORD	flg = (info.flags & USE_OSCACHE_READ) ? FILE_FLAG_SEQUENTIAL_SCAN : (FILE_FLAG_NO_BUFFERING|FILE_FLAG_SEQUENTIAL_SCAN);

	while (1) {
		if (!::BackupRead(stat->hFile, (LPBYTE)&sid, STRMID_OFFSET, &size, FALSE, TRUE, &context) || size != STRMID_OFFSET) {
			DWORD	errCode = ::GetLastError();
			if (errCode == 0 || errCode == ERROR_INVALID_HANDLE)
				ret = TRUE;
			else if (info.flags & (REPORT_ACL_ERROR|REPORT_STREAM_ERROR))
				ConfirmErr("BackupRead(head)", src);
			break;
		}
		if (sid.dwStreamNameSize) {
			if (!::BackupRead(stat->hFile, streamName, sid.dwStreamNameSize, &size, FALSE, TRUE, &context)) {
				if (info.flags & REPORT_STREAM_ERROR)
					ConfirmErr("BackupRead(name)", src);
				break;
			}
			lSetCharV((LPBYTE)streamName + sid.dwStreamNameSize, 0, 0);	// terminate Ȃ߁idwStreamNameSize ̓oCgj
		}

		if (sid.dwStreamId == BACKUP_ALTERNATE_DATA && enableStream) {
			FileStat *subStat = (FileStat *)(fileStatBuf.Buf() + fileStatBuf.UsedSize());

			if (++altdata_cnt >= MAX_ALTSTREAM) {
				if (info.flags & REPORT_STREAM_ERROR)
					ConfirmErr("Too Many AltStream", src);
				break;
			}
			openFiles[openFilesCnt++] = subStat;
			subStat->fileID = nextFileID++;
			subStat->hFile = INVALID_HANDLE_VALUE;
			subStat->nFileSizeLow = sid.Size.LowPart;
			subStat->nFileSizeHigh = sid.Size.HighPart;
			subStat->dwFileAttributes = 0;	// ALTSTREAM
			subStat->renameCount = 0;
			subStat->lastError = 0;
			subStat->size = sid.dwStreamNameSize + offsetof(FileStat, cFileName);
			subStat->minSize = ALIGN_SIZE(subStat->size, 8);

			fileStatBuf.AddUsedSize(subStat->minSize);
			if (fileStatBuf.RemainSize() <= maxStatSize && fileStatBuf.Grow(MIN_ATTR_BUF) == FALSE) {
				ConfirmErr("Can't alloc memory(fileStatBuf2)", NULL, FALSE);
				break;
			}
			memcpy(subStat->cFileName, streamName, size + CHAR_LEN_V);
			memcpy(MakeAddr(src, src_len), subStat->cFileName, size + CHAR_LEN_V);

			if ((subStat->hFile = CreateFileV(src, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, flg, 0)) == INVALID_HANDLE_VALUE) {
				if (info.flags & REPORT_STREAM_ERROR)
					ConfirmErr("OpenFile(Stream)", src);
				subStat->lastError = ::GetLastError();
				break;
			}
		}
		else if ((sid.dwStreamId == BACKUP_SECURITY_DATA || sid.dwStreamId == BACKUP_EA_DATA) && enableAcl) {
			BYTE	*&data		= sid.dwStreamId == BACKUP_SECURITY_DATA ? stat->acl     : stat->ead;
			DWORD	&data_size	= sid.dwStreamId == BACKUP_SECURITY_DATA ? stat->aclSize : stat->eadSize;
			if (data || sid.Size.HighPart) {	// łɊi[ς
				if (info.flags & REPORT_ACL_ERROR)
					ConfirmErr("Duplicate or Too big ACL/EADATA", src);
				break;
			}
			data = fileStatBuf.Buf() + fileStatBuf.UsedSize();
			data_size = sid.Size.LowPart + STRMID_OFFSET;
			fileStatBuf.AddUsedSize(ALIGN_SIZE(data_size, 8));
			if (fileStatBuf.RemainSize() <= maxStatSize && !fileStatBuf.Grow(ALIGN_SIZE(maxStatSize, MIN_ATTR_BUF))) {
				if (info.flags & REPORT_ACL_ERROR)
					ConfirmErr("Can't alloc memory(fileStat(ACL/EADATA))", src, FALSE);
				break;
			}
			memcpy(data, &sid, STRMID_OFFSET);
			if (!::BackupRead(stat->hFile, data + STRMID_OFFSET, sid.Size.LowPart, &size, FALSE, TRUE, &context) || size <= 0) {
				if (info.flags & REPORT_ACL_ERROR)
					ConfirmErr("BackupRead(ACL/EADATA)", src);
				break;
			}
			sid.Size.LowPart = 0;
		}
		if ((sid.Size.LowPart || sid.Size.HighPart) && !::BackupSeek(stat->hFile, sid.Size.LowPart, sid.Size.HighPart, &lowSeek, &highSeek, &context)) {
			if (info.flags & (REPORT_ACL_ERROR|REPORT_STREAM_ERROR))
				ConfirmErr("BackupSeek", src);
			break;
		}
	}
	::BackupRead(stat->hFile, 0, 0, 0, TRUE, FALSE, &context);

	if (ret)
		total.readAclStream++;
	else
		total.errAclStream++;

	return	ret;
}

BOOL FastCopy::ReadMultiFilesProc(void)
{
	for (int i=0; !isAbort && i < openFilesCnt; ) {
		ReadFileProc(i, &i);
	}
	return	!isAbort;
}

BOOL FastCopy::CloseMultiFilesProc(void)
{
	if ((info.flags & LISTING_ONLY) == 0) {
		for (int i=0; i < openFilesCnt; i++) {
			if (openFiles[i]->hFile != INVALID_HANDLE_VALUE)
				::CloseHandle(openFiles[i]->hFile);
		}
	}
	openFilesCnt = 0;

	return	!isAbort;
}

BOOL FastCopy::ReadFileProc(int start_idx, int *end_idx)
{
	BOOL		ret = TRUE;
	FileStat	*stat = openFiles[start_idx];
	_int64		file_size = stat->FileSize();
	DWORD		trans_size;
	ReqBuf		req_buf;
	Command		command = enableAcl || enableStream ? stat->dwFileAttributes ? WRITE_BACKUP_FILE : WRITE_BACKUP_ALTSTREAM : WRITE_FILE;
	BOOL		isStream = command == WRITE_BACKUP_ALTSTREAM;
	int			&totalFiles = isStream ? total.readAclStream : total.readFiles;
	int			&totalErrFiles = isStream ? total.errAclStream  : total.errFiles;

	*end_idx = start_idx + 1;

	if (file_size == 0 || (info.flags & LISTING_ONLY)) {
		ret = SendRequest(command, 0, stat);
		if (command != WRITE_BACKUP_FILE || (info.flags & LISTING_ONLY))
			return	ret;
	}

	if (stat->hFile == INVALID_HANDLE_VALUE) {
		::SetLastError(stat->lastError);
		totalErrFiles++;
		if (isStream && !(info.flags & REPORT_STREAM_ERROR) || ConfirmErr(isStream ? "OpenFile(Stream)" : "OpenFile", MakeAddr(src, srcPrefixLen)) == Confirm::CONTINUE_RESULT) {
			stat->SetFileSize(0);
			if (info.flags & CREATE_OPENERR_FILE)
				SendRequest(command, 0, stat);
		}
		return	FALSE;
	}

	::SetFilePointer(stat->hFile, 0, NULL, FILE_BEGIN);

	for (_int64 remain_size=file_size; remain_size > 0; remain_size -= trans_size) {
		BOOL	isFirst = remain_size == file_size;
		if ((ret = PrepareReqBuf(offsetof(ReqHeader, stat) + (isFirst ? stat->minSize : 0), remain_size, stat->fileID, &req_buf)) == FALSE)
			break;
		if ((ret = ::ReadFile(stat->hFile, req_buf.buf, req_buf.bufSize, &trans_size, NULL)) == FALSE || trans_size == 0) {
			totalErrFiles++;
			if (isStream && !(info.flags & REPORT_STREAM_ERROR) || ConfirmErr(isStream ? "ReadFile(stream" : ret && !trans_size ? "ReadFile(truncate)" : "ReadFile", MakeAddr(src, srcPrefixLen)) == Confirm::CONTINUE_RESULT || !isFirst) {
				stat->SetFileSize(0);
				req_buf.bufSize = 0;
				if ((info.flags & CREATE_OPENERR_FILE) || !isFirst) {
					SendRequest(isFirst ? command : WRITE_ABORT, &req_buf, isFirst ? stat : 0);
				}
			}
			break;
		}
		total.readTrans += trans_size;
		SendRequest(isFirst ? command : WRITE_FILE_CONT, &req_buf, isFirst ? stat : 0);
		if (waitTick && remain_size > trans_size) Wait();
	}

	if (command == WRITE_BACKUP_FILE) {
		while (ret && !isAbort && (*end_idx) < openFilesCnt && openFiles[(*end_idx)]->dwFileAttributes == 0) {
			ret = ReadFileProc(*end_idx, end_idx);
		}
		if (ret && stat->acl) {
			if ((ret = PrepareReqBuf(offsetof(ReqHeader, stat) + stat->minSize, stat->aclSize, stat->fileID, &req_buf))) {
				memcpy(req_buf.buf, stat->acl, stat->aclSize);
				ret = SendRequest(WRITE_BACKUP_ACL, &req_buf, stat);
			}
		}
		if (ret && stat->ead) {
			if ((ret = PrepareReqBuf(offsetof(ReqHeader, stat) + stat->minSize, stat->eadSize, stat->fileID, &req_buf))) {
				memcpy(req_buf.buf, stat->ead, stat->eadSize);
				ret = SendRequest(WRITE_BACKUP_EADATA, &req_buf, stat);
			}
		}
		SendRequest(WRITE_BACKUP_END, 0, 0);
	}

	return	ret && !isAbort;
}

void FastCopy::ReadDestStatRequest(void)
{
	cv.Lock();
	readDestStatQueue = TRUE;
	cv.Notify();
	cv.UnLock();
}

BOOL FastCopy::WaitReadDestStat(void)
{
	cv.Lock();
	while (readDestStatQueue && !isAbort)
		cv.Wait();
	cv.UnLock();
	return	readDestStatResult;
}

BOOL FastCopy::ReadDestStat(void)
{
	HANDLE		fh;
	int			num = 0, len;
	FileStat	*dstStat, **dstStatIdx;
	WIN32_FIND_DATAW	fdat;

	if (!isSameDrv)
		cv.UnLock();
	readDestStatResult = TRUE;
	dstStat = (FileStat *)dstStatBuf.Buf();
	dstStatIdx = (FileStat **)dstStatIdxBuf.Buf();
	dstStatBuf.SetUsedSize(0);
	dstStatIdxBuf.SetUsedSize(0);

	if ((fh = FindFirstFileV(confirmDst, &fdat)) == INVALID_HANDLE_VALUE) {
		if (::GetLastError() != ERROR_FILE_NOT_FOUND && strcmpV(MakeAddr(confirmDst, dstBaseLen), ASTERISK_V) == 0) {
			readDestStatResult = FALSE;
			total.errDirs++;
			ConfirmErr("FindFirstFile(stat)", MakeAddr(confirmDst, dstPrefixLen));
		}	// t@Cw肵ẴRs[ŁARs[悪Ȃꍇ́A
			// GgȂł̐Ƃ݂Ȃ
		goto END;
	}
	do {
		if (IsParentOrSelfDirs(fdat.cFileName))
			continue;
		dstStatIdx[num++] = dstStat;
		len = FdatToFileStat(&fdat, dstStat, TRUE);
		dstStatBuf.AddUsedSize(len);

		//  stat p buffer ̃Zbg
		dstStat = (FileStat *)(dstStatBuf.Buf() + dstStatBuf.UsedSize());
		dstStatIdxBuf.AddUsedSize(sizeof(FileStat *));

		if (dstStatBuf.RemainSize() <= maxStatSize && dstStatBuf.Grow(MIN_ATTR_BUF) == FALSE) {
			ConfirmErr("Can't alloc memory(dstStatBuf)", NULL, FALSE);
			break;
		}
		if (dstStatIdxBuf.RemainSize() <= sizeof(FileStat *) && dstStatIdxBuf.Grow(MIN_ATTRIDX_BUF) == FALSE) {
			ConfirmErr("Can't alloc memory(dstStatIdxBuf)", NULL, FALSE);
			break;
		}
	}
	while (!isAbort && FindNextFileV(fh, &fdat));

	if (!isAbort && ::GetLastError() != ERROR_NO_MORE_FILES) {
		total.errFiles++;
		readDestStatResult = FALSE;
		ConfirmErr("FindNextFile(stat)", MakeAddr(confirmDst, dstPrefixLen));
	}
	::FindClose(fh);

END:
	if (readDestStatResult)
		readDestStatResult = MakeHashTable();

	if (!isSameDrv) {
		cv.Lock();
		cv.Notify();
	}
	readDestStatQueue = FALSE;
	return	readDestStatResult;
}

BOOL FastCopy::MakeHashTable(void)
{
	int		num = dstStatIdxBuf.UsedSize() / sizeof(FileStat *);
	int		require_size = hash.RequireSize(num), grow_size;

	if ((grow_size = require_size - dstStatIdxBuf.RemainSize()) > 0 && dstStatIdxBuf.Grow(ALIGN_SIZE(grow_size, MIN_ATTRIDX_BUF)) == FALSE) {
		ConfirmErr("Can't alloc memory(dstStatIdxBuf2)", NULL, FALSE);
		return	FALSE;
	}

	return	hash.Init((FileStat **)dstStatIdxBuf.Buf(), num, dstStatIdxBuf.Buf() + dstStatIdxBuf.UsedSize());
}

int StatHash::HashNum(int data_num)
{
	return	data_num | 1;
}

BOOL StatHash::Init(FileStat **data, int data_num, void *tbl_buf)
{
	hashNum = HashNum(data_num);
	hashTbl = (FileStat **)tbl_buf;
	memset(hashTbl, 0, hashNum * sizeof(FileStat *));

	for (int i=0; i < data_num; i++) {
		for (FileStat **stat = hashTbl+(data[i]->hashVal % hashNum); *stat; stat = &(*stat)->next)
			;
		*stat = data[i];
	}
	return	TRUE;
}

FileStat *StatHash::Search(void *upperName, DWORD hash_val)
{
	for (FileStat *target = hashTbl[hash_val % hashNum]; target; target = target->next) {
		if (target->hashVal == hash_val && strcmpV(target->upperName, upperName) == 0)
			return	target;
	}

	return	NULL;
}

/*=========================================================================
     F DeleteThread
  T  v F DELETE_MODE 
     F 
     F 
=========================================================================*/
unsigned WINAPI FastCopy::DeleteThread(void *fastCopyObj)
{
	return	((FastCopy *)fastCopyObj)->DeleteThreadCore();
}

BOOL FastCopy::DeleteThreadCore(void)
{
	if ((info.flags & PRE_SEARCH) && info.mode == DELETE_MODE)
		PreSearch();

	for (int i=0; i < srcArray.Num() && !isAbort; i++) {
		if (InitDeletePath(i))
			DeleteProc(dst, dstBaseLen);
	}
	FinishNotify();
	return	TRUE;
}

/*
	폜
*/
BOOL FastCopy::DeleteProc(void *path, int dir_len)
{
	HANDLE		hDir;
	BOOL		ret = TRUE;
	FileStat	stat;
	WIN32_FIND_DATAW fdat;

	if ((hDir = FindFirstFileV(path, &fdat)) == INVALID_HANDLE_VALUE) {
		total.errDirs++;
		return	ConfirmErr("FindFirstFile(del)", MakeAddr(path, dstPrefixLen)), FALSE;
	}

	do {
		if (IsParentOrSelfDirs(fdat.cFileName))
			continue;

		if (waitTick) Wait((waitTick + 9) / 10);

		// 폜w肵fBNǵi[gjłȂAtB^[OΏ
		if (isFilter && (dstBaseLen != dir_len || isMetaSrc) && !FilterCheck(path, fdat.cFileName, fdat.dwFileAttributes)) {
			total.filterSkips++;
			continue;
		}

		stat.nFileSizeLow		= fdat.nFileSizeLow;
		stat.nFileSizeHigh		= fdat.nFileSizeHigh;
		stat.dwFileAttributes	= fdat.dwFileAttributes;

		if (stat.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			ret = DeleteDirProc(path, dir_len, fdat.cFileName, &stat);
		else
			ret = DeleteFileProc(path, dir_len, fdat.cFileName, &stat);
	}
	while (!isAbort && FindNextFileV(hDir, &fdat));

	if (!isAbort && ret && ::GetLastError() != ERROR_NO_MORE_FILES) {
		ret = FALSE;
		ConfirmErr("FindNextFile(del)", MakeAddr(path, dstPrefixLen));
	}

	::FindClose(hDir);

	return	ret && !isAbort;
}

BOOL FastCopy::DeleteDirProc(void *path, int dir_len, void *fname, FileStat *stat)
{
	int		new_dir_len = dir_len + sprintfV(MakeAddr(path, dir_len), FMT_CAT_ASTER_V, fname) -1;
	BOOL	ret;
	int		cur_skips = total.filterSkips;

	if (info.mode == DELETE_MODE && (info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA))) {
		memcpy(MakeAddr(confirmDst, dir_len), MakeAddr(path, dir_len), (new_dir_len - dir_len + 2) * CHAR_LEN_V);
	}

	if ((ret = DeleteProc(path, new_dir_len)), isAbort)
		return	ret;

	if (isFilter && (cur_skips != total.filterSkips || regExp[INC_EXP][FILE_EXP].IsRegistered()))
		return	ret;

	SetChar(path, new_dir_len - 1, 0);

	if (info.flags & LISTING_ONLY) {
		PutList(MakeAddr(path, dstPrefixLen), PL_DIRECTORY|PL_DELETE);
	}
	else {
		if (stat->dwFileAttributes & FILE_ATTRIBUTE_READONLY)
			SetFileAttributesV(path, stat->dwFileAttributes & ~FILE_ATTRIBUTE_READONLY);

		if (info.mode == DELETE_MODE && (info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA))) {
			if (RenameRandomFname(dst, confirmDst, dir_len, new_dir_len-dir_len-1)) {
				path = confirmDst;
			}
		}

		if ((ret = RemoveDirectoryV(path)) == FALSE) {
			total.errDirs++;
			return	ConfirmErr("RemoveDirectory", MakeAddr(path, dstPrefixLen)), FALSE;
		}
	}

	total.deleteDirs++;
	return	ret;
}

BOOL FastCopy::DeleteFileProc(void *path, int dir_len, void *fname, FileStat *stat)
{
	int	len = sprintfV(MakeAddr(path, dir_len), FMT_STR_V, fname);

	if (info.flags & LISTING_ONLY) {
		PutList(MakeAddr(path, dstPrefixLen), PL_DELETE);
	}
	else {
		if (stat->dwFileAttributes & FILE_ATTRIBUTE_READONLY)
			SetFileAttributesV(path, FILE_ATTRIBUTE_NORMAL);

		if (info.mode == DELETE_MODE && (info.flags & (OVERWRITE_DELETE|OVERWRITE_DELETE_NSA))) {
			if (RenameRandomFname(dst, confirmDst, dir_len, len)) {
				path = confirmDst;
			}
			if (stat->FileSize()) {
				if (WriteRandomData(path, stat, TRUE) == FALSE) {
					total.errFiles++;
					return	ConfirmErr("OverWrite", MakeAddr(path, dstPrefixLen)), FALSE;
				}
				total.writeFiles++;
			}
		}

		if (DeleteFileV(path) == FALSE) {
			total.errFiles++;
			return	ConfirmErr("DeleteFile", MakeAddr(path, dstPrefixLen)), FALSE;
		}
	}

	total.deleteFiles++;
	total.deleteTrans += stat->FileSize();
	return	TRUE;
}


/*=========================================================================
  T  v F ㏑폜p[`
=========================================================================*/
void FastCopy::SetupRandomDataBuf(void)
{
	RandomDataBuf	*data = (RandomDataBuf *)mainBuf.Buf();

	data->is_nsa = (info.flags & OVERWRITE_DELETE_NSA) ? TRUE : FALSE;
	data->base_size = max(PAGE_SIZE, dstSectorSize);
	data->buf_size = mainBuf.Size() - data->base_size;
	data->buf[0] = mainBuf.Buf() + data->base_size;

	if (data->is_nsa) {
		data->buf_size /= 3;
		data->buf_size = (data->buf_size / data->base_size) * data->base_size;
		data->buf_size = min(info.maxTransSize, data->buf_size);

		data->buf[1] = data->buf[0] + data->buf_size;
		data->buf[2] = data->buf[1] + data->buf_size;
		for (int i=0, max=data->buf_size / sizeof(int) * 2; i < max; i++) {	// buf[0], buf[1] 
			*((int *)data->buf[0] + i) = rand();	//  CryptGenRandom() or MT gp邩
		}
		memset(data->buf[2], 0, data->buf_size);
	}
	else {
		data->buf_size = min(info.maxTransSize, data->buf_size);

		for (int i=0, max=data->buf_size / sizeof(int); i < max; i++) {
			*((int *)data->buf[0] + i) = rand();	//  CryptGenRandom() or MT gp邩
		}
	}
}

void FastCopy::GenRandomName(void *path, int fname_len, int ext_len)
{
	static char *char_dict = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

	for (int i=0; i < fname_len; i++) {
		SetChar(path, i, char_dict[(rand() >> 4) % 62]);
	}
	if (ext_len) {
		SetChar(path, fname_len - ext_len, '.');
	}
	SetChar(path, fname_len, 0);
}

BOOL FastCopy::RenameRandomFname(void *org_path, void *rename_path, int dir_len, int fname_len)
{
	void	*fname = MakeAddr(org_path, dir_len);
	void	*rename_fname = MakeAddr(rename_path, dir_len);
	void	*dot = strrchrV(fname, '.');
	int		ext_len = dot ? fname_len - DiffLen(dot, fname) : 0;

	for (int i=fname_len; i <= MAX_FNAME_LEN; i++) {
		for (int j=0; j < 128; j++) {
			GenRandomName(rename_fname, i, ext_len);
			if (MoveFileV(org_path, rename_path)) {
				return	TRUE;
			}
			else if (::GetLastError() != ERROR_ALREADY_EXISTS) {
				return	FALSE;
			}
		}
	}
	return	FALSE;
}

BOOL FastCopy::WriteRandomData(void *path, FileStat *stat, BOOL is_truncate)
{
	BOOL	isNonBuf = dstFsType != FSTYPE_NETWORK && (stat->FileSize() >= max(nbMinSize, PAGE_SIZE) || (stat->FileSize() % dstSectorSize) == 0) && (info.flags & USE_OSCACHE_WRITE) == 0 ? TRUE : FALSE;
	DWORD	flg = (isNonBuf ? (FILE_FLAG_SEQUENTIAL_SCAN|FILE_FLAG_NO_BUFFERING) : FILE_FLAG_SEQUENTIAL_SCAN);
	DWORD	trans_size;
	HANDLE	hFile = CreateFileV(path, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, flg, 0);
	BOOL	ret = TRUE;

	if (hFile == INVALID_HANDLE_VALUE) {
		return	ConfirmErr("Write by Random Data(open)", MakeAddr(path, dstPrefixLen)), FALSE;
	}

	RandomDataBuf	*data = (RandomDataBuf *)mainBuf.Buf();
	_int64			file_size = isNonBuf ? ALIGN_SIZE(stat->FileSize(), dstSectorSize) : stat->FileSize();

	for (int i=0, end=data->is_nsa ? 3 : 1; i < end && ret; i++) {
		::SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
		for (_int64 remain_size=file_size; remain_size > 0; remain_size -= trans_size) {
			if (!(ret = ::WriteFile(hFile, data->buf[i], (DWORD)(remain_size > data->buf_size ? data->buf_size : remain_size), &trans_size, NULL))) {
				break;
			}
			total.writeTrans += trans_size;
		}
		total.writeTrans -= file_size - stat->FileSize();
		::FlushFileBuffers(hFile);
	}

	if (is_truncate) {
		::SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
		::SetEndOfFile(hFile);
	}

	::CloseHandle(hFile);
	return	ret;
}

/*=========================================================================
     F WriteThread
  T  v F Write 
     F 
     F 
=========================================================================*/
unsigned WINAPI FastCopy::WriteThread(void *fastCopyObj)
{
	return	((FastCopy *)fastCopyObj)->WriteThreadCore();
}

BOOL FastCopy::WriteThreadCore(void)
{
	BOOL	ret = WriteProc(dstBaseLen);

	if (info.mode == MOVE_MODE && !isAbort && errBuf.UsedSize() == 0 && total.errFiles == 0 && total.errDirs == 0)
		DeleteThreadCore();	// G[ȂȂA\[X̍폜
	else
		FinishNotify();

	return	ret;
}

BOOL FastCopy::FinishNotify(void)
{
	endTick = ::GetTickCount();
	return	::PostMessage(info.notifyWnd->hWnd, info.uNotifyMsg, END_NOTIFY, 0);
}

BOOL FastCopy::WriteProc(int dir_len)
{
	BOOL		ret = TRUE;
	int			new_dir_len;
	HANDLE		fh;
	FileStat	sv_stat;

	while (!isAbort) {
		if ((ret = RecvRequest()) == FALSE || writeReq->command == REQ_EOF) {
			break;
		}
		if (writeReq->command == WRITE_FILE || writeReq->command == WRITE_BACKUP_FILE) {
			if (writeReq->stat.renameCount == 0)
				new_dir_len = dir_len + sprintfV(MakeAddr(dst, dir_len), FMT_STR_V, writeReq->stat.cFileName);
			else
				new_dir_len = dir_len + MakeRenameName(MakeAddr(dst, dir_len), writeReq->stat.renameCount, writeReq->stat.cFileName);

			if (mkdirQueueBuf.UsedSize())
				ExecDirQueue();

			if (info.flags & LISTING_ONLY) {
				PutList(MakeAddr(dst, dstPrefixLen), PL_NORMAL);
				total.writeFiles++;
				total.writeTrans += writeReq->stat.FileSize();
			}
			else if ((ret = WriteFileProc(new_dir_len)), isAbort)
				break;
		}
		else if (writeReq->command == MKDIR || writeReq->command == INTODIR) {
			memcpy(&sv_stat, &writeReq->stat, offsetof(FileStat, cFileName));

			if (writeReq->stat.renameCount == 0)
				new_dir_len = dir_len + sprintfV(MakeAddr(dst, dir_len), FMT_STR_V, writeReq->stat.cFileName);
			else
				new_dir_len = dir_len + MakeRenameName(MakeAddr(dst, dir_len), writeReq->stat.renameCount, writeReq->stat.cFileName);

			if (sv_stat.acl) {
				if (mkdirAclBuf.RemainSize() < (int)sv_stat.aclSize && !mkdirAclBuf.Grow(ALIGN_SIZE(sv_stat.aclSize, MIN_MKDIRACL_BUF))) {
					ConfirmErr("Can't alloc memory(mkdirAclBuf)", NULL, FALSE);
					break;
				}
				memcpy(mkdirAclBuf.Buf() + mkdirAclBuf.UsedSize(), sv_stat.acl, sv_stat.aclSize);
				sv_stat.acl = mkdirAclBuf.Buf() + mkdirAclBuf.UsedSize();
				mkdirAclBuf.AddUsedSize(sv_stat.aclSize);
			}

			if (writeReq->command == MKDIR) {
				if (info.flags & SKIP_EMPTYDIR) {
					if (mkdirQueueBuf.RemainSize() < sizeof(int) && mkdirQueueBuf.Grow(MIN_MKDIRQUEUE_BUF) == FALSE) {
						ConfirmErr("Can't alloc memory(mkdirQueueBuf)", NULL, FALSE);
						break;
					}
					*(int *)(mkdirQueueBuf.Buf() + mkdirQueueBuf.UsedSize()) = new_dir_len;
					mkdirQueueBuf.AddUsedSize(sizeof(int));
				}
				else {
					if (info.flags & LISTING_ONLY)
						PutList(MakeAddr(dst, dstPrefixLen), PL_DIRECTORY);
					else
						CreateDirectoryV(dst, NULL);
					total.writeDirs++;
				}
			}
			strcpyV(MakeAddr(dst, new_dir_len), BACK_SLASH_V);

			if ((ret = WriteProc(new_dir_len + 1)), isAbort)	// ċA
				break;

			// ^CX^v/ACL/̃Zbg
			if (writeReq->command != MKDIR || (info.flags & SKIP_EMPTYDIR) == 0 || mkdirQueueBuf.UsedSize() == 0) {
				SetChar(dst, new_dir_len, 0);	//  '\\' 
				if ((info.flags & LISTING_ONLY) == 0) {
					if (sv_stat.dwFileAttributes & (FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_SYSTEM))
						SetFileAttributesV(dst, sv_stat.dwFileAttributes);

					DWORD	mode = GENERIC_WRITE | (sv_stat.acl && enableAcl ? (WRITE_OWNER|WRITE_DAC) : 0);

					if (IS_WINNT_V && (fh = CreateFileV(dst, mode, 0, 0, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0)) != INVALID_HANDLE_VALUE) {
						if (sv_stat.acl) {
							void	*backupContent = NULL;
							DWORD	size;
							if (!(ret = ::BackupWrite(fh, sv_stat.acl, sv_stat.aclSize, &size, FALSE, TRUE, &backupContent))) {
								if (info.flags & REPORT_ACL_ERROR)
									ConfirmErr("BackupWrite(DIRACL)", MakeAddr(dst, dstPrefixLen));
							}
							::BackupWrite(fh, NULL, NULL, NULL, TRUE, TRUE, &backupContent);
						}
						::SetFileTime(fh, &sv_stat.ftCreationTime, &sv_stat.ftLastAccessTime, &sv_stat.ftLastWriteTime);
						::CloseHandle(fh);
					}
				}
			}
			if (sv_stat.acl) {
				mkdirAclBuf.AddUsedSize(-(int)sv_stat.aclSize);
			}
			if (mkdirQueueBuf.UsedSize()) {
				mkdirQueueBuf.AddUsedSize(-(int)sizeof(int));
			}
		}
		else if (writeReq->command == DELETE_FILES) {
			if (writeReq->stat.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				ret = DeleteDirProc(dst, dir_len, writeReq->stat.cFileName, &writeReq->stat);
			else
				ret = DeleteFileProc(dst, dir_len, writeReq->stat.cFileName, &writeReq->stat);
		}
		else if (writeReq->command == RETURN_PARENT) {
			break;
		}
		else if (writeReq->command < WRITE_FILE && writeReq->command > REQ_EOF) {
			ret = FALSE;
			ConfirmErr("Illegal Request (internal error)", NULL, FALSE);
			break;
		}
	}
	return	ret && !isAbort;
}

BOOL FastCopy::ExecDirQueue(void)
{
	for (int offset=0; offset < mkdirQueueBuf.UsedSize(); offset += sizeof(int)) {
		int dir_len = *(int *)(mkdirQueueBuf.Buf() + offset);
		SetChar(dst, dir_len, 0);
		if (info.flags & LISTING_ONLY)
			PutList(MakeAddr(dst, dstPrefixLen), PL_DIRECTORY);
		else
			CreateDirectoryV(dst, NULL);
		SetChar(dst, dir_len, '\\');
		total.writeDirs++;
	}
	mkdirQueueBuf.SetUsedSize(0);
	return	TRUE;
}


BOOL FastCopy::WriteFileProc(int dst_len)
{
	HANDLE		fh, fh2;
	BOOL		ret = TRUE;
	_int64		file_size = writeReq->stat.FileSize();
	_int64		remain = file_size;
	DWORD		trans_size;
	FileStat	*stat = &writeReq->stat, sv_stat;
	BOOL		isNonBuf = dstFsType != FSTYPE_NETWORK && (file_size >= nbMinSize || (file_size % dstSectorSize) == 0) && (info.flags & USE_OSCACHE_WRITE) == 0 ? TRUE : FALSE;
	BOOL		isReOpen = isNonBuf && (file_size % dstSectorSize) ? TRUE : FALSE;
	Command		command = writeReq->command;
	BOOL		isStream = command == WRITE_BACKUP_ALTSTREAM;
	int			&totalFiles = isStream ? total.writeAclStream : total.writeFiles;
	int			&totalErrFiles = isStream ? total.errAclStream  : total.errFiles;

	// writeReq  stat ҔāA𗘗p
	if (command == WRITE_BACKUP_FILE || file_size > writeReq->bufSize) {
		memcpy((stat = &sv_stat), &writeReq->stat, offsetof(FileStat, cFileName));
	}

	DWORD	mode = GENERIC_WRITE;
	DWORD	flg = (isNonBuf ? (FILE_FLAG_SEQUENTIAL_SCAN|FILE_FLAG_NO_BUFFERING) : FILE_FLAG_SEQUENTIAL_SCAN);

	if (command == WRITE_BACKUP_FILE) {
		mode |= WRITE_OWNER|WRITE_DAC;
		flg  |= FILE_FLAG_BACKUP_SEMANTICS;
	}

	if (waitTick) Wait((waitTick + 9) / 10);

	if ((fh = CreateFileV(dst, mode, FILE_SHARE_WRITE, 0, CREATE_ALWAYS, flg, 0)) == INVALID_HANDLE_VALUE) {
		SetFileAttributesV(dst, FILE_ATTRIBUTE_NORMAL);
		fh = CreateFileV(dst, mode, FILE_SHARE_WRITE, 0, CREATE_ALWAYS, flg, 0);
	}

	if (fh == INVALID_HANDLE_VALUE) {
		SetErrFileID(stat->fileID);
		totalErrFiles++;
		if (!isStream || (info.flags & REPORT_STREAM_ERROR))
			ConfirmErr(isStream ? "CreateFile(stream)" : "CreateFile", MakeAddr(dst, dstPrefixLen));
		return	FALSE;
	}
	if (isReOpen) {
		flg &= ~FILE_FLAG_NO_BUFFERING;
		fh2 = CreateFileV(dst, mode, FILE_SHARE_WRITE, 0, OPEN_EXISTING, flg, 0);
	}

	if (file_size > writeReq->bufSize) {
		_int64	alloc_size = isNonBuf ? ALIGN_SIZE(file_size, dstSectorSize) : file_size;
		LONG	high_size = (LONG)(alloc_size >> 32);
		::SetFilePointer(fh, (LONG)alloc_size, &high_size, FILE_BEGIN);
		::SetEndOfFile(fh);
		::SetFilePointer(fh, 0, NULL, FILE_BEGIN);
	}

	while (remain > 0) {
		if (!(ret = ::WriteFile(fh, writeReq->buf, remain >= writeReq->bufSize ? writeReq->bufSize : (DWORD)(isNonBuf ? ALIGN_SIZE(remain, dstSectorSize) : remain), &trans_size, NULL))) {
			SetErrFileID(stat->fileID);
			if (!isStream || (info.flags & REPORT_STREAM_ERROR))
				ConfirmErr(isStream ? "WriteFile(stream)" : "WriteFile", MakeAddr(dst, dstPrefixLen), GetLastError() != ERROR_DISK_FULL || isStream);
			break;
		}
		if ((remain -= trans_size) > 0) {	// 
			total.writeTrans += trans_size;
			if (RecvRequest() == FALSE || writeReq->command != WRITE_FILE_CONT) {
				ret = FALSE;
				remain = 0;
				break;
			}
			if (waitTick) Wait();
		}
		else {
			total.writeTrans += trans_size + remain;
		}
	}

	if (isReOpen) {
		::CloseHandle(fh);
		if (fh2 == INVALID_HANDLE_VALUE) {
			for (int i=0; i < 10 && !isAbort; i++) {	// ECX`FbN\tg΍
				total.openRetry++;

				if ((fh2 = CreateFileV(dst, mode, FILE_SHARE_WRITE, 0, OPEN_EXISTING, flg, 0)) != INVALID_HANDLE_VALUE)
					break;
				if (::GetLastError() != ERROR_SHARING_VIOLATION)
					break;
				::Sleep(i * i * 10);
			}
			if (fh2 == INVALID_HANDLE_VALUE) {
				totalErrFiles++;
				if (isStream || (info.flags & REPORT_STREAM_ERROR))
					ConfirmErr(isStream ? "CreateFile2(stream)" : "CreateFile2", MakeAddr(dst, dstPrefixLen));
				return	FALSE;
			}
		}
		fh = fh2;
	}

	if (ret && remain != 0) {
		::SetFilePointer(fh, stat->nFileSizeLow, (LONG *)&stat->nFileSizeHigh, FILE_BEGIN);
		if ((ret = ::SetEndOfFile(fh)) == FALSE) {
			if (!isStream || (info.flags & REPORT_STREAM_ERROR))
				ConfirmErr(isStream ? "SetEndOfFile(stream)" : "SetEndOfFile", MakeAddr(dst, dstPrefixLen));
		}
	}

	if (ret && command == WRITE_BACKUP_FILE) {
		/* ret = */ WriteFileBackupProc(fh, dst_len);	// ACL/EADATA/STREAM G[͖
	}
	if (ret && !isStream) {
		::SetFileTime(fh, &stat->ftCreationTime, &stat->ftLastAccessTime, &stat->ftLastWriteTime);
	}

	::CloseHandle(fh);

	if (ret) {
		if (!isStream && (stat->dwFileAttributes & (FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_SYSTEM)))
			SetFileAttributesV(dst, stat->dwFileAttributes);
		totalFiles++;
	}
	else {
		totalErrFiles;
		DeleteFileV(dst);
	}
	return	ret && !isAbort;
}

BOOL FastCopy::WriteFileBackupProc(HANDLE fh, int dst_len)
{
	BOOL	ret;
	DWORD	size;
	void	*backupContent = NULL;	// for BackupWrite
	dst_len = lstrlenV(dst);

	while ((ret = RecvRequest()) && writeReq->command != WRITE_BACKUP_END) {
		switch (writeReq->command) {
		case WRITE_BACKUP_ACL: case WRITE_BACKUP_EADATA:
			SetLastError(0);
			if (!(ret = ::BackupWrite(fh, writeReq->buf, writeReq->command == WRITE_BACKUP_ACL ? writeReq->stat.aclSize : writeReq->stat.eadSize, &size, FALSE, TRUE, &backupContent))) {
				if (info.flags & REPORT_ACL_ERROR)
					ConfirmErr("BackupWrite(ACL/EADATA)", MakeAddr(dst, dstPrefixLen));
			}
			break;

		case WRITE_BACKUP_ALTSTREAM:
			strcpyV(MakeAddr(dst, dst_len), writeReq->stat.cFileName);
			ret = WriteFileProc(0);
			lSetCharV(dst, dst_len, 0);
			break;

		default:
			ret = FALSE;
			break;
		}
	}
	if (backupContent) {
		::BackupWrite(fh, NULL, NULL, NULL, TRUE, TRUE, &backupContent);
	}
	return	ret;
}

BOOL FastCopy::ChangeToWriteModeCore(void)
{
	writeReqList.MoveList(&readReqList);
	cv.Notify();
	while (!writeReqList.IsEmpty() && !isAbort)
		cv.Wait();
	return	!isAbort;
}

BOOL FastCopy::ChangeToWriteMode(void)
{
	cv.Lock();
	BOOL	ret = ChangeToWriteModeCore();
	cv.UnLock();
	return	ret;
}

BOOL FastCopy::AllocReqBuf(int req_size, _int64 _data_size, ReqBuf *buf)
{
	int		max_free = mainBuf.Buf() + mainBuf.Size() - usedOffset;
	int		max_trans = waitTick && (info.flags & AUTOSLOW_IOLIMIT) ? MIN_BUF : maxTransSize;
	int		data_size = (_data_size > max_trans) ? max_trans : (int)_data_size;
	int		align_data_size = ALIGN_SIZE(data_size, 8);
	int		sector_data_size = ALIGN_SIZE(data_size, sectorSize);
	int		require_size = align_data_size + req_size;
	BYTE	*align_offset = data_size ? (BYTE *)ALIGN_SIZE((u_int)usedOffset, sectorSize) : usedOffset;

	max_free -= align_offset - usedOffset;

	if (data_size && align_data_size + req_size < sector_data_size)
		require_size = sector_data_size;

	if (require_size > max_free) {
		if (max_free < MIN_BUF) {
			align_offset = mainBuf.Buf();
			if (isSameDrv && ChangeToWriteModeCore() == FALSE)	// Read -> Write ؂ւ
				return	FALSE;
		}
		else {
			data_size = max_free - req_size;
			align_data_size = data_size = (data_size / BIGTRANS_ALIGN) * BIGTRANS_ALIGN;
			require_size = data_size + req_size;
		}
	}
	buf->buf = align_offset;
	buf->bufSize = ALIGN_SIZE(data_size, sectorSize);
	buf->req = (ReqHeader *)(buf->buf + align_data_size);
	buf->reqSize = req_size;

	while (!writeReqList.IsEmpty() && !isAbort) {	// isSameDrv == TRUE ̏ꍇAK Empty
		if (buf->buf == mainBuf.Buf()) {
			if (freeOffset < usedOffset && (freeOffset - mainBuf.Buf()) >= require_size)
				break;
		}
		else {
			if (freeOffset < usedOffset || buf->buf + require_size <= freeOffset)
				break;
		}
		cv.Wait();
	}

	usedOffset = buf->buf + require_size;
	return	!isAbort;
}

BOOL FastCopy::PrepareReqBuf(int req_size, _int64 data_size, _int64 file_id, ReqBuf *buf)
{
	BOOL ret = TRUE;

	cv.Lock();

	if (errFileID) {
		if (errFileID == file_id)
			ret = FALSE;
		errFileID = 0;
	}

	if (ret)
		ret = AllocReqBuf(req_size, data_size, buf);

	cv.UnLock();

	return	ret;
}

BOOL FastCopy::SendRequest(Command command, ReqBuf *buf, FileStat *stat)
{
	BOOL	ret = TRUE;
	ReqBuf	tmp_buf;

	cv.Lock();

	if (buf == NULL) {
		buf = &tmp_buf;
		ret = AllocReqBuf(offsetof(ReqHeader, stat) + (stat ? stat->minSize : 0), 0, buf);
	}

	if (ret && !isAbort) {
		ReqHeader	*readReq;
		readReq				= buf->req;
		readReq->reqSize	= buf->reqSize;
		readReq->command	= command;
		readReq->buf		= buf->buf;
		readReq->bufSize	= buf->bufSize;
		if (stat)
			memcpy(&readReq->stat, stat, stat->minSize);

		if (isSameDrv) {
			readReqList.AddObj(readReq);
		}
		else {
			writeReqList.AddObj(readReq);
			cv.Notify();
		}
	}

	cv.UnLock();

	return	ret && !isAbort;
}

BOOL FastCopy::RecvRequest(void)
{
	cv.Lock();

	if (writeReq)
		WriteReqDone();

	if (readDestStatQueue)
		ReadDestStat();

	while (writeReqList.IsEmpty() && !isAbort) {
		cv.Wait();
		if (readDestStatQueue)
			ReadDestStat();
	}
	writeReq = writeReqList.TopObj();

	if (writeReq && writeReq->command == REQ_EOF)
		WriteReqDone();

	cv.UnLock();

	return	writeReq && !isAbort;
}

void FastCopy::WriteReqDone(void)
{
	writeReqList.DelObj(writeReq);
	freeOffset = (BYTE *)writeReq + writeReq->reqSize;
	if (!isSameDrv || writeReqList.IsEmpty())
		cv.Notify();
}

void FastCopy::SetErrFileID(_int64 file_id)
{
	cv.Lock();
	errFileID = file_id;
	cv.UnLock();
}

BOOL FastCopy::End(void)
{
	isAbort = TRUE;

	while (hWriteThread || hReadThread) { // WaitForMultipleObject() ւ̕ύX͌
		cv.Lock();
		cv.Notify();
		cv.UnLock();

		if (hReadThread && ::WaitForSingleObject(hReadThread, 100) != WAIT_TIMEOUT) {
			::CloseHandle(hReadThread);
			hReadThread = NULL;
		}
		if (hWriteThread && ::WaitForSingleObject(hWriteThread, 100) != WAIT_TIMEOUT) {
			::CloseHandle(hWriteThread);
			hWriteThread = NULL;
		}
	}

	::ReleaseMutex(hRunMutex);

	delete [] openFiles;
	openFiles = NULL;

	mainBuf.FreeBuf();
	mkdirAclBuf.FreeBuf();
	mkdirQueueBuf.FreeBuf();
	dstStatIdxBuf.FreeBuf();
	dstStatBuf.FreeBuf();
	dirStatBuf.FreeBuf();
	fileStatBuf.FreeBuf();
	listBuf.FreeBuf();
	errBuf.FreeBuf();

	return	TRUE;
}

BOOL FastCopy::Suspend(void)
{
	if (!hReadThread && !hWriteThread || isSuspend)
		return	FALSE;

	if (hReadThread)
		::SuspendThread(hReadThread);

	if (hWriteThread)
		::SuspendThread(hWriteThread);

	isSuspend = TRUE;
	suspendTick = ::GetTickCount();

	return	TRUE;
}

BOOL FastCopy::Resume(void)
{
	if (!hReadThread && !hWriteThread || !isSuspend)
		return	FALSE;

	isSuspend = FALSE;
	startTick += (::GetTickCount() - suspendTick);

	if (hReadThread)
		::ResumeThread(hReadThread);

	if (hWriteThread)
		::ResumeThread(hWriteThread);

	return	TRUE;
}

BOOL FastCopy::GetTransInfo(TransInfo *ti, BOOL fullInfo)
{
	ti->total = total;
	ti->listBuf = &listBuf;
	ti->listCs = &listCs;
	ti->errBuf = &errBuf;
	ti->errCs = &errCs;
	ti->isSameDrv = isSameDrv;
	ti->ignoreErr = info.ignoreErr;
	ti->waitTick = waitTick;
	ti->tickCount = (isSuspend ? suspendTick : endTick ? endTick : ::GetTickCount()) - startTick;
	if (fullInfo) {
		if (IS_WINNT_V)
			ConvertExternalPath(MakeAddr(dst, dstPrefixLen), ti->curPath, sizeof(ti->curPath)/sizeof(WCHAR));
		else
			sprintfV(ti->curPath, "%.250s", dst);
	}
	return	TRUE;
}

int FastCopy::FdatToFileStat(WIN32_FIND_DATAW *fdat, FileStat *stat, BOOL is_usehash)
{
	stat->fileID			= 0;
	stat->ftCreationTime	= fdat->ftCreationTime;
	stat->ftLastAccessTime	= fdat->ftLastAccessTime;
	stat->ftLastWriteTime	= fdat->ftLastWriteTime;
	stat->nFileSizeLow		= fdat->nFileSizeLow;
	stat->nFileSizeHigh		= fdat->nFileSizeHigh;
	stat->dwFileAttributes	= fdat->dwFileAttributes;
	stat->hFile				= INVALID_HANDLE_VALUE;
	stat->lastError			= 0;
	stat->isExists			= FALSE;
	stat->renameCount		= 0;
	stat->acl				= NULL;
	stat->aclSize			= 0;
	stat->ead				= NULL;
	stat->eadSize			= 0;
	stat->next				= NULL;

	int	len = (sprintfV(stat->cFileName, FMT_STR_V, fdat->cFileName) + 1) * CHAR_LEN_V;
	stat->size = len + offsetof(FileStat, cFileName);
	stat->minSize = ALIGN_SIZE(stat->size, 8);

	if (is_usehash) {
		stat->upperName = stat->cFileName + len;
		memcpy(stat->upperName, stat->cFileName, len);
		CharUpperV(stat->upperName);
		stat->hashVal = hashVal.MakeHash(stat->upperName, len);
		stat->size += len;
	}
	stat->size = ALIGN_SIZE(stat->size, 8);

	return	stat->size;
}

BOOL FastCopy::ConvertExternalPath(const void *path, void *buf, int max_buf, BOOL to_ansi)
{
	if (GetChar(path, 0) == '\\' && GetChar(path, 1) != '\\') {	// UNC
		if (to_ansi) {
			*(char *)buf = '\\';
			buf = (char *)buf + 1;
		}
		else {
			SetChar(buf, 0, '\\');
			buf = MakeAddr(buf, 1);
		}
		max_buf--;
	}
	if (to_ansi)
		::WideCharToMultiByte(CP_ACP, 0, (WCHAR *)path, -1, (char *)buf, max_buf, 0, 0);
	else
		sprintfV(buf, L"%.*s", max_buf, path);
	return	TRUE;
}

FastCopy::Confirm::Result FastCopy::ConfirmErr(const char *message, const void *path, BOOL allow_continue)
{
	if (isAbort)
		return	Confirm::CANCEL_RESULT;

	char	path_buf[MAX_PATH_EX], msg_buf[MAX_PATH_EX + 100];
	DWORD	err_code = ::GetLastError();

	if (path && IS_WINNT_V) {
		ConvertExternalPath(path, path_buf, sizeof(path_buf), TRUE);
		path = path_buf;
	}

	int	len = sprintf(msg_buf, "%s(", message);

	len += FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS, NULL, err_code, info.lcid > 0 ? LANGIDFROMLCID(info.lcid) : MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), msg_buf + len, MAX_PATH_EX, NULL);
	if (msg_buf[len-1] == '\n')
		len -= 2;

	len += sprintf(msg_buf + len, "%u) : %s", err_code, path ? path : "");

	WriteErrLog(msg_buf, len);

	if (allow_continue == FALSE)
		isAbort = TRUE;
	else if (info.ignoreErr)
		return	Confirm::CONTINUE_RESULT;

	isSuspend = TRUE;
	suspendTick = ::GetTickCount();

	Confirm	 confirm = { msg_buf, allow_continue, path, Confirm::CONTINUE_RESULT };
	info.notifyWnd->SendMessage(info.uNotifyMsg, CONFIRM_NOTIFY, (LPARAM)&confirm);

	isSuspend = FALSE;
	startTick += (::GetTickCount() - suspendTick);

	switch (confirm.result) {
	case Confirm::IGNORE_RESULT:
		info.ignoreErr = TRUE;
		confirm.result = Confirm::CONTINUE_RESULT;
		break;

	case Confirm::CANCEL_RESULT:
		isAbort = TRUE;
		break;
	}
	return	confirm.result;
}

BOOL FastCopy::WriteErrLog(char *message, int len)
{
#define ERRMSG_SIZE		30
	::EnterCriticalSection(&errCs);

	BOOL	ret = TRUE;
	char	*msg_buf = (char *)errBuf.Buf() + errBuf.UsedSize();

	if (len == -1)
		len = strlen(message);
	len += ERRMSG_SIZE;

	if (errBuf.UsedSize() + len <= errBuf.MaxSize()) {
		if (errBuf.RemainSize() > len || errBuf.Grow(ALIGN_SIZE(len, PAGE_SIZE))) {
			len = sprintf(msg_buf, "%s%s", errBuf.UsedSize() ? "\r\n" : "", message);
			errBuf.AddUsedSize(len);
		}
	}
	else {
		if (errBuf.RemainSize()) {
			sprintf(msg_buf, "%s", "\r\n Too Many Errors...");
			errBuf.SetUsedSize(errBuf.MaxSize());
		}
		ret = FALSE;
	}
	::LeaveCriticalSection(&errCs);
	return	ret;
}

BOOL FastCopy::Wait(DWORD tick)
{
#define SLEEP_UNIT	200
	int	svWaitTick = (int)waitTick;
	int	remain = (int)(tick ? tick : waitTick);

	while (remain > 0 && !isAbort) {
		int	unit = remain > SLEEP_UNIT ? SLEEP_UNIT : remain;
		::Sleep(unit);
		int	curWaitTick = waitTick;
		if (curWaitTick != SUSPEND_WAITTICK) {
			remain -= unit + (svWaitTick - curWaitTick);
			svWaitTick = curWaitTick;
		}
	}
	return	!isAbort;
}

